import random
from .engine import Value
from typing import List, Union

class Neuron:
    def __init__(self, n_in: int):
        self.w = [Value(random.uniform(-1, 1)) for _ in range(n_in)]
        self.b = Value(random.uniform(-1, 1))

    def __call__(self, x) -> Value:
        # w * x + b
        from functools import reduce
        from operator import add
        
        if len(x) != len(self.w):
            raise ValueError(f"mismatch dimension: x: {len(x)}, w: {len(self.w)}")
        act = reduce(add, [w_i * x_i for w_i, x_i in zip(self.w, x)]) + self.b
        return act.tanh()
    def parameters(self):
        return self.w + [self.b]

class Layer:
    def __init__(self, n_in: int, n_out: int):
        self.neurons = [Neuron(n_in) for _ in range(n_out)]

    def __call__(self, x) -> Union[List[Value], Value]:
        out = [n(x) for n in self.neurons]
        return out[0] if len(out) == 1 else out

    def parameters(self):
        out = []
        for neuron in self.neurons:
            for p in neuron.parameters():
                out.append(p)
        return out


class MLP:
    def __init__(self, n_in: int, n_outs: List[int]):
        sizes = [n_in] + n_outs
        self.layers = []
        for i in range(len(n_outs)):
            self.layers.append(Layer(sizes[i], sizes[i+1]))

    def __call__(self, x):
        for layer in self.layers:
            x = layer(x)
        return x

    def parameters(self):
        out = []
        for layer in self.layers:
            for l in layer.parameters():
                out.append(l)
        return out
